"""
clana is a toolkit for classifier analysis.

It specifies some file formats and comes with some tools for typical tasks of
classifier analysis.
"""
# Core Library
import logging.config
import os
import random
from typing import Optional

# Third party
import click
import matplotlib

# First party
import clana
import clana.distribution
import clana.get_cm
import clana.get_cm_simple
import clana.visualize_cm

matplotlib.use("Agg")


config = clana.utils.load_cfg(verbose=True)
logging.config.dictConfig(config["LOGGING"])
logging.getLogger("matplotlib").setLevel("WARN")
random.seed(0)


@click.group()
@click.version_option(version=clana.__version__)
def entry_point() -> None:
    """
    Clana is a toolkit for classifier analysis.

    See https://arxiv.org/abs/1707.09725, Chapter 4.
    """


gt_option = click.option(
    "--gt",
    "gt_filepath",
    required=True,
    type=click.Path(exists=True),
    help="CSV file with delimiter ;",
)
predictions_option = click.option(
    "--predictions",
    "predictions_filepath",
    required=True,
    type=click.Path(exists=True),
    help="CSV file with delimiter ;",
)


@entry_point.group()
def get_cm() -> None:
    """Generate a confusion matrix file."""


@get_cm.command(name="simple")
@click.option(
    "--labels",
    "label_filepath",
    required=True,
    type=click.Path(exists=True),
    help="CSV file with delimiter ;",
)
@predictions_option
@gt_option
@click.option(
    "--clean",
    default=False,
    is_flag=True,
    help="Remove classes that the classifier doesn't know",
)
def get_cm_simple(
    label_filepath: str, predictions_filepath: str, gt_filepath: str, clean: bool
) -> None:
    """
    Generate a confusion matrix.

    The input can be a flat list of predictions and a flat list of ground truth
    elements. Each prediction is on its own line. Additional information can be
    after a semicolon.
    """
    clana.get_cm_simple.main(label_filepath, gt_filepath, predictions_filepath, clean)


@get_cm.command(name="standard")
@predictions_option
@gt_option
@click.option("--n", "n", required=True, type=int, help="Number of classes")
def get_cm_standard(predictions_filepath: str, gt_filepath: str, n: int) -> None:
    """
    Generate a confusion matrix from predictions and ground truth.

    The predictions need to be a list of `identifier;prediction` and the
    ground truth needs to be a list of `identifier;truth` of same length.
    """
    clana.get_cm.main(predictions_filepath, gt_filepath, n)


@entry_point.command(name="distribution")
@gt_option
def distribution(gt_filepath: str) -> None:
    """Get the distribution of classes in a dataset."""
    clana.distribution.main(gt_filepath)


@entry_point.command(name="visualize")
@click.option("--cm", "cm_file", type=click.Path(exists=True), required=True)
@click.option(
    "--perm",
    "perm_file",
    help="json file which defines a permutation to start with.",
    type=click.Path(),
    default=None,
)
@click.option(
    "--steps",
    default=1000,
    show_default=True,
    help="Number of steps to find a good permutation.",
)
@click.option("--labels", "labels_file", default="")
@click.option(
    "--zero_diagonal",
    is_flag=True,
    help=(
        "Good classifiers have the highest elements on the diagonal. "
        "This option sets the diagonal to zero so that errors "
        "can be seen more easily."
    ),
)
@click.option(
    "--output",
    "output_image_path",
    type=click.Path(exists=False),
    help="Where to store the image (either .png or .pdf)",
    default=os.path.abspath(config["visualize"]["save_path"]),
    show_default=True,
)
@click.option(
    "--limit_classes", type=int, help="Limit the number of classes in the output"
)
def visualize(
    cm_file: str,
    perm_file: str,
    steps: int,
    labels_file: str,
    zero_diagonal: bool,
    output_image_path: str,
    limit_classes: Optional[int] = None,
) -> None:
    """Optimize and visualize a confusion matrix."""
    print_file_format_issues(cm_file, labels_file, perm_file)
    clana.visualize_cm.main(
        cm_file,
        perm_file,
        steps,
        labels_file,
        zero_diagonal,
        limit_classes,
        output_image_path,
    )


def print_file_format_issues(cm_file: str, labels_file: str, perm_file: str) -> None:
    """
    Get all potential issues of the file formats.

    Parameters
    ----------
    cm_file : str
    labels_file : str
    perm_file : str
    """
    if not (cm_file.lower().endswith("json") or cm_file.lower().endswith("csv")):
        print(f"[WARNING] A json file is expected for the cm_file, but was {cm_file}")
    if not (perm_file is None or perm_file.lower().endswith("json")):
        print(
            f"[WARNING] A json file is expected fo the perm_file, but was {perm_file}"
        )
    cm = clana.io.read_confusion_matrix(cm_file)
    labels = clana.io.read_labels(labels_file, len(cm))
    special_labels = ["UNK"]
    if len(labels) - len(special_labels) < len(cm):
        print(
            "[WARNING] The shape of the confusion matrix is {cm_shape}, but "
            "only {nb_labels} labels were found: {labels}".format(
                cm_shape=cm.shape, nb_labels=len(labels), labels=labels
            )
        )
        print(
            "Please keep in mind that the first row of the labels file is "
            "the header of the CSV (delimiter: ;)"
        )



"""Optimize the column order of a confusion matrix."""

# Core Library
import logging
import random
from typing import Callable, List, NamedTuple, Tuple, Union

# Third party
import numpy as np
import numpy.typing as npt

logger = logging.getLogger(__name__)


class OptimizationResult(NamedTuple):
    """The result of a matrix column/row order optimiataion (CMO)."""

    cm: npt.NDArray
    perm: npt.NDArray


def calculate_score(cm: npt.NDArray, weights: npt.NDArray) -> int:
    """
    Calculate a score how close big elements of cm are to the diagonal.

    Parameters
    ----------
    cm : npt.NDArray
        The confusion matrix
    weights : npt.NDArray
        The weights matrix.
        It has to have the same shape as the confusion matrix

    Examples
    --------
    >>> cm = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    >>> weights = calculate_weight_matrix(3)
    >>> weights.shape
    (3, 3)
    >>> calculate_score(cm, weights)
    32
    """
    assert cm.shape == weights.shape
    return int(np.tensordot(cm, weights, axes=((0, 1), (0, 1))))


def simulated_annealing(
    current_cm: npt.NDArray,
    current_perm: Union[None, List[int], npt.NDArray] = None,
    score: Callable[[npt.NDArray, npt.NDArray], float] = calculate_score,
    steps: int = 2 * 10**5,
    temp: float = 100.0,
    cooling_factor: float = 0.99,
    deterministic: bool = False,
) -> OptimizationResult:
    """
    Optimize current_cm by randomly swapping elements.

    Parameters
    ----------
    current_cm : npt.NDArray
    current_perm : None or iterable, optional (default: None)
    score: Callable[[npt.NDArray, npt.NDArray], float], optional
        (default: )
    steps : int, optional (default: 2 * 10**4)
    temp : float > 0.0, optional (default: 100.0)
        Temperature
    cooling_factor: float in (0, 1), optional (default: 0.99)

    Returns
    -------
    best_result : OptimizationResult
    """
    if temp <= 0.0:
        raise ValueError(f"temp={temp} needs to be positive")
    if cooling_factor <= 0.0 or cooling_factor >= 1.0:
        raise ValueError(
            f"cooling_factor={cooling_factor} needs to be in the interval (0, 1)"
        )
    n = len(current_cm)
    logger.info(f"n={n}")

    # Load the initial permutation
    if current_perm is None:
        current_perm = np.array(list(range(n)))
    current_perm = np.array(current_perm)

    # Pre-calculate weights
    weights = calculate_weight_matrix(n)

    # Apply the permutation
    current_cm = apply_permutation(current_cm, current_perm)
    current_score = score(current_cm, weights)

    best_cm = current_cm
    best_score = current_score
    best_perm = current_perm

    logger.info(f"## Starting Score: {current_score:0.2f}")
    for step in range(steps):
        tmp_cm = np.array(current_cm, copy=True)
        perm, make_swap = generate_permutation(n, current_perm, tmp_cm)
        tmp_score = score(tmp_cm, weights)

        # Should be swapped?
        if deterministic:
            chance = 1.0
        else:
            chance = random.random()
            temp *= 0.99
        hot_prob_thresh = min(1, np.exp(-(tmp_score - current_score) / temp))
        if chance <= hot_prob_thresh:
            changed = False
            if best_score > tmp_score:  # minimize
                best_perm = perm
                best_cm = tmp_cm
                best_score = tmp_score
                changed = True
            current_score = tmp_score
            current_cm = tmp_cm
            current_perm = perm
            if changed:
                logger.info(
                    (
                        "Current: %0.2f (best: %0.2f, "
                        "hot_prob_thresh=%0.4f%%, step=%i, swap=%s)"
                    ),
                    current_score,
                    best_score,
                    (hot_prob_thresh * 100),
                    step,
                    str(make_swap),
                )
    return OptimizationResult(cm=best_cm, perm=best_perm)


def calculate_weight_matrix(n: int) -> npt.NDArray:
    """
    Calculate the weights for each position.

    The weight is the distance to the diagonal.

    Parameters
    ----------
    n : int

    Examples
    --------
    >>> calculate_weight_matrix(3)
    array([[0.  , 1.01, 2.02],
           [1.01, 0.  , 1.03],
           [2.02, 1.03, 0.  ]])
    """
    weights = np.abs(np.arange(n) - np.arange(n)[:, None])
    weights = np.array(weights, dtype=float)
    for i in range(n):
        for j in range(n):
            if i == j:
                continue
            weights[i][j] += (i + j) * 0.01
    return weights


def generate_permutation(
    n: int, current_perm: npt.NDArray, tmp_cm: npt.NDArray
) -> Tuple[npt.NDArray, bool]:
    """
    Generate a new permutation.

    Parameters
    ----------
    n : int
    current_perm : List[int]
    tmp_cm : npt.NDArray

    Return
    ------
    perm, make_swap : List[int], bool
    """
    swap_prob = 0.5
    make_swap = random.random() < swap_prob
    if n < 3:
        # In this case block-swaps don't make any sense
        make_swap = True
    if make_swap:
        # Choose what to swap
        i = random.randint(0, n - 1)
        j = i
        while j == i:
            j = random.randint(0, n - 1)
        # Define permutation
        perm = swap_1d(current_perm.copy(), i, j)
        # Define values after swap
        tmp_cm = swap(tmp_cm, i, j)
    else:
        # block-swap
        block_len = n
        while block_len >= n - 1:
            from_start = random.randint(0, n - 3)
            from_end = random.randint(from_start + 1, n - 2)
            block_len = from_start - from_end
        insert_pos = from_start
        while not (insert_pos < from_start or insert_pos > from_end):
            insert_pos = random.randint(0, n - 1)
        perm = move_1d(current_perm.copy(), from_start, from_end, insert_pos)

        # Define values after swap
        tmp_cm = move(tmp_cm, from_start, from_end, insert_pos)
    return perm, make_swap


def apply_permutation(
    cm: npt.NDArray, perm: Union[List[int], npt.NDArray]
) -> npt.NDArray:
    """
    Apply permutation to a matrix.

    Parameters
    ----------
    cm : ndarray
    perm : List[int]

    Examples
    --------
    >>> cm = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    >>> perm = np.array([2, 0, 1])
    >>> apply_permutation(cm, perm)
    array([[8, 6, 7],
           [2, 0, 1],
           [5, 3, 4]])
    """
    return cm[perm].transpose()[perm].transpose()


def move_1d(
    perm: npt.NDArray, from_start: int, from_end: int, insert_pos: int
) -> npt.NDArray:
    """
    Move a block in a list.

    Parameters
    ----------
    perm : npt.NDArray
        Permutation
    from_start : int
    from_end : int
    insert_pos : int

    Returns
    -------
    perm : npt.NDArray
        The new permutation
    """
    if not (insert_pos < from_start or insert_pos > from_end):
        raise ValueError(
            f"insert_pos={insert_pos} needs to be smaller than "
            f"from_start={from_start} or greater than from_end={from_end}"
        )
    if insert_pos > from_end:
        p_new = list(range(from_end + 1, insert_pos + 1)) + list(
            range(from_start, from_end + 1)
        )
    else:
        p_new = list(range(from_start, from_end + 1)) + list(
            range(insert_pos, from_start)
        )
    p_old = sorted(p_new)
    perm[p_old] = perm[p_new]
    return perm


def move(
    cm: npt.NDArray, from_start: int, from_end: int, insert_pos: int
) -> npt.NDArray:
    """
    Move rows from_start - from_end to insert_pos in-place.

    Parameters
    ----------
    cm : npt.NDArray
    from_start : int
    from_end : int
    insert_pos : int

    Returns
    -------
    cm : npt.NDArray

    Examples
    --------
    >>> cm = np.array([[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 0, 1], [2, 3, 4, 5]])
    >>> move(cm, 1, 2, 0)
    array([[5, 6, 4, 7],
           [9, 0, 8, 1],
           [1, 2, 0, 3],
           [3, 4, 2, 5]])
    """
    if not (insert_pos < from_start or insert_pos > from_end):
        raise ValueError(
            f"insert_pos={insert_pos} needs to be smaller than "
            f"from_start={from_start} or greater than from_end={from_end}"
        )
    if insert_pos > from_end:
        p_new = list(range(from_end + 1, insert_pos + 1)) + list(
            range(from_start, from_end + 1)
        )
    else:
        p_new = list(range(from_start, from_end + 1)) + list(
            range(insert_pos, from_start)
        )
    p_old = sorted(p_new)
    # swap columns
    cm[:, p_old] = cm[:, p_new]
    # swap rows
    cm[p_old, :] = cm[p_new, :]
    return cm


def swap_1d(perm: npt.NDArray, i: int, j: int) -> npt.NDArray:
    """
    Swap two elements of a 1-D numpy array in-place.

    Parameters
    ----------
    parm : npt.NDArray
    i : int
    j : int

    Examples
    --------
    >>> perm = np.array([2, 1, 2, 3, 4, 5, 6])
    >>> swap_1d(perm, 2, 6)
    array([2, 1, 6, 3, 4, 5, 2])
    """
    perm[i], perm[j] = perm[j], perm[i]
    return perm


def swap(cm: npt.NDArray, i: int, j: int) -> npt.NDArray:
    """
    Swap row and column i and j in-place.

    Parameters
    ----------
    cm : npt.NDArray
    i : int
    j : int

    Returns
    -------
    cm : npt.NDArray

    Examples
    --------
    >>> cm = np.array([[0, 1, 2], [3, 4, 5], [6, 7, 8]])
    >>> swap(cm, 2, 0)
    array([[8, 7, 6],
           [5, 4, 3],
           [2, 1, 0]])
    """
    # swap columns
    copy = cm[:, i].copy()
    cm[:, i] = cm[:, j]
    cm[:, j] = copy
    # swap rows
    copy = cm[i, :].copy()
    cm[i, :] = cm[j, :]
    cm[j, :] = copy
    return cm



"""Utility functions for clana."""

# Core Library
import csv
import os
from typing import Any, Dict, List, Optional

# Third party
import yaml
from pkg_resources import resource_filename


def load_labels(labels_file: str, n: int) -> List[str]:
    """
    Load labels from a CSV file.

    Parameters
    ----------
    labels_file : str
    n : int

    Returns
    -------
    labels : List[str]
    """
    if n < 0:
        raise ValueError(f"n={n} needs to be non-negative")
    if os.path.isfile(labels_file):
        # Read CSV file
        with open(labels_file) as fp:
            reader = csv.reader(fp, delimiter=";", quotechar='"')
            next(reader, None)  # skip the headers
            parsed_csv = list(reader)
            labels = [el[0] for el in parsed_csv]  # short by default
    else:
        labels = [str(el) for el in range(n)]
    return labels


def load_cfg(
    yaml_filepath: Optional[str] = None, verbose: bool = False
) -> Dict[str, Any]:
    """
    Load a YAML configuration file.

    Parameters
    ----------
    yaml_filepath : str, optional (default: package config file)

    Returns
    -------
    cfg : Dict[str, Any]
    """
    if yaml_filepath is None:
        yaml_filepath = resource_filename("clana", "config.yaml")
    # Read YAML experiment definition file
    if verbose:
        print(f"Load config from {yaml_filepath}...")
    with open(yaml_filepath) as stream:
        cfg = yaml.safe_load(stream)
    cfg = make_paths_absolute(os.path.dirname(yaml_filepath), cfg)
    return cfg


def make_paths_absolute(dir_: str, cfg: Dict[str, Any]) -> Dict[str, Any]:
    """
    Make all values for keys ending with `_path` absolute to dir_.

    Parameters
    ----------
    dir_ : str
    cfg : Dict[str, Any]

    Returns
    -------
    cfg : Dict[str, Any]
    """
    for key in cfg.keys():
        if hasattr(key, "endswith") and key.endswith("_path"):
            if cfg[key].startswith("~"):
                cfg[key] = os.path.expanduser(cfg[key])
            else:
                cfg[key] = os.path.join(dir_, cfg[key])
            cfg[key] = os.path.abspath(cfg[key])
        if type(cfg[key]) is dict:
            cfg[key] = make_paths_absolute(dir_, cfg[key])
    return cfg



"""Get the version."""

# Third party
import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("clana").version
except pkg_resources.DistributionNotFound:
    __version__ = "not installed"



"""
Optimize confusion matrix.

For more information, see

* http://cs.stackexchange.com/q/70627/2914
* http://datascience.stackexchange.com/q/17079/8820
"""

# Core Library
import json
import logging
from typing import List, Optional, Tuple

# Third party
import numpy as np
import numpy.typing as npt
from jinja2 import Template
from mpl_toolkits.axes_grid1 import make_axes_locatable
from pkg_resources import resource_filename
from sklearn.metrics import silhouette_score

# First party
import clana.clustering
import clana.cm_metrics
import clana.io
import clana.utils
from clana.optimize import (
    calculate_score,
    calculate_weight_matrix,
    simulated_annealing,
)

cfg = clana.utils.load_cfg()
logger = logging.getLogger(__name__)


def main(
    cm_file: str,
    perm_file: str,
    steps: int,
    labels_file: str,
    zero_diagonal: bool,
    limit_classes: Optional[int] = None,
    output: Optional[str] = None,
) -> None:
    """
    Run optimization and generate output.

    Parameters
    ----------
    cm_file : str
    perm_file : str
    steps : int
    labels_file : str
    zero_diagonal : bool
    limit_classes : int, optional (default: no limit)
    output : str
    """
    cm = clana.io.read_confusion_matrix(cm_file)
    perm = clana.io.read_permutation(cm_file, perm_file)
    labels = clana.io.read_labels(labels_file, len(cm))
    n, m = cm.shape
    if n != m:
        raise ValueError(
            f"Confusion matrix is expected to be square, but was {n} x {m}"
        )
    if len(labels) - 1 != n:
        print(
            "Confusion matrix is {n} x {n}, but len(labels)={nb_labels}".format(
                n=n, nb_labels=len(labels)
            )
        )

    cm_orig = cm.copy()

    get_cm_problems(cm, labels)

    weights = calculate_weight_matrix(len(cm))
    print(f"Score: {calculate_score(cm, weights)}")
    result = simulated_annealing(
        cm, perm, score=calculate_score, deterministic=True, steps=steps
    )
    print(f"Score: {calculate_score(result.cm, weights)}")
    print(f"Perm: {list(result.perm)}")
    clana.io.ClanaCfg.store_permutation(cm_file, result.perm, steps)
    labels = [labels[i] for i in result.perm]
    class_indices = list(range(len(labels)))
    class_indices = [class_indices[i] for i in result.perm]
    logger.info(f"Classes: {labels}")
    acc = clana.cm_metrics.get_accuracy(cm_orig)
    print(f"Accuracy: {acc * 100:0.2f}%")
    start = 0
    if limit_classes is None:
        limit_classes = len(cm)
    if output is None:
        output = cfg["visualize"]["save_path"]
    plot_cm(
        result.cm[start:limit_classes, start:limit_classes],
        zero_diagonal=zero_diagonal,
        labels=labels[start:limit_classes],
        output=output,
    )
    create_html_cm(
        result.cm[start:limit_classes, start:limit_classes],
        zero_diagonal=zero_diagonal,
        labels=labels[start:limit_classes],
    )
    if len(cm) < 5:
        print(
            f"You only have {len(cm)} classes. Clustering for less than "
            "5 classes should be done manually."
        )
        return
    grouping = clana.clustering.extract_clusters(result.cm, labels)
    y_pred = [0]
    cluster_i = 0
    for el in grouping:
        if el:
            cluster_i += 1
        y_pred.append(cluster_i)
    logger.info(f"silhouette_score={silhouette_score(cm, y_pred)}")
    # Store grouping as hierarchy
    with open(cfg["visualize"]["hierarchy_path"], "w") as outfile:
        hierarchy = clana.clustering.apply_grouping(class_indices, grouping)
        hierarchy_mixed = clana.clustering._remove_single_element_groups(hierarchy)
        str_ = json.dumps(
            hierarchy_mixed,
            indent=4,
            sort_keys=True,
            separators=(",", ":"),
            ensure_ascii=False,
        )
        outfile.write(str_)

    # Print nice
    for group in clana.clustering.apply_grouping(labels, grouping):
        print(f"\t{len(group)}: {list(group)}")


def get_cm_problems(cm: npt.NDArray, labels: List[str]) -> None:
    """
    Find problems of a classifier by analzing its confusion matrix.

    Parameters
    ----------
    cm : ndarray
    labels : List[str]
    """
    n = len(cm)

    # Find classes which are not present in the dataset
    for i in range(n):
        if sum(cm[i]) == 0:
            logger.warning(f"The class '{labels[i]}' was not in the dataset.")

    # Find classes which are never predicted
    cm = cm.transpose()
    never_predicted = []
    for i in range(n):
        if sum(cm[i]) == 0:
            never_predicted.append(labels[i])
    if len(never_predicted) > 0:
        logger.warning(f"The following classes were never predicted: {never_predicted}")


def plot_cm(
    cm: npt.NDArray,
    zero_diagonal: bool = False,
    labels: Optional[List[str]] = None,
    output: str = cfg["visualize"]["save_path"],
) -> None:
    """
    Plot a confusion matrix.

    Parameters
    ----------
    cm : npt.NDArray
    zero_diagonal : bool, optional (default: False)
    labels : Optional[List[str]]
        If this is not given, then numbers are assigned to the classes
    """
    # Third party
    from matplotlib import pyplot as plt
    from matplotlib.colors import LogNorm

    n = len(cm)
    if zero_diagonal:
        for i in range(n):
            cm[i][i] = 0
    if n > 20:
        size = int(n / 4.0)
    else:
        size = 5
    fig = plt.figure(figsize=(size, size), dpi=80)
    plt.clf()
    ax = fig.add_subplot(111)
    ax.set_aspect(1)
    if labels is None:
        labels = [str(i) for i in range(len(cm))]
    x = list(range(len(cm)))
    plt.xticks(x, labels, rotation=cfg["visualize"]["xlabels_rotation"])
    y = list(range(len(cm)))
    plt.yticks(y, labels, rotation=cfg["visualize"]["ylabels_rotation"])
    if cfg["visualize"]["norm"] == "LogNorm":
        norm = LogNorm(vmin=max(1, np.min(cm)), vmax=np.max(cm))  # type: ignore
    elif cfg["visualize"]["norm"] is None:
        norm = None
    else:
        raise NotImplementedError(
            f"visualize->norm={cfg['visualize']['norm']} is not implemented. "
            "Try None or LogNorm"
        )
    res = ax.imshow(
        np.array(cm),
        cmap=cfg["visualize"]["colormap"],
        interpolation=cfg["visualize"]["interpolation"],
        norm=norm,
    )
    width, height = cm.shape

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.5)
    plt.colorbar(res, cax=cax)
    plt.tight_layout()

    logger.info(f"Save figure at '{output}'")
    plt.savefig(output)


def create_html_cm(
    cm: npt.NDArray, zero_diagonal: bool = False, labels: Optional[List[str]] = None
) -> None:
    """
    Plot a confusion matrix.

    Parameters
    ----------
    cm : npt.NDArray
    zero_diagonal : bool, optional (default: False)
        If this is set to True, then the diagonal is overwritten with zeroes.
    labels : Optional[List[str]]
        If this is not given, then numbers are assigned to the classes
    """
    if labels is None:
        labels = [str(i) for i in range(len(cm))]

    el_max = 200

    template_path = resource_filename("clana", "templates/base.html")
    with open(template_path) as f:
        base = f.read()

    cm_t = cm.transpose()
    header_cells = []
    for i, label in enumerate(labels):
        precision = cm[i][i] / float(sum(cm_t[i]))
        background_color = "transparent"
        if precision < 0.2:
            background_color = "red"
        elif precision > 0.98:
            background_color = "green"
        header_cells.append(
            {
                "precision": f"{precision:0.2f}",
                "background-color": background_color,
                "label": label,
            }
        )

    body_rows = []
    for i, label, row in zip(range(len(labels)), labels, cm):
        body_row = []
        row_str = [str(el) for el in row]
        support = sum(row)
        recall = cm[i][i] / float(support)
        background_color = "transparent"
        if recall < 0.2:
            background_color = "red"
        elif recall >= 0.98:
            background_color = "green"
        body_row.append(
            {
                "label": label,
                "recall": f"{recall:.2f}",
                "background-color": background_color,
            }
        )
        for _j, pred_label, el in zip(range(len(labels)), labels, row_str):
            background_color = "transparent"
            if el == "0":
                el = ""
            else:
                background_color = get_color_code(float(el), el_max)

            body_row.append(
                {
                    "label": el,
                    "true": label,
                    "pred": pred_label,
                    "background-color": background_color,
                }
            )

        body_rows.append({"row": body_row, "support": support})

    html_template = Template(base)
    html = html_template.render(header_cells=header_cells, body_rows=body_rows)

    with open(cfg["visualize"]["html_save_path"], "w") as f:
        f.write(html)


def get_color(white_to_black: float) -> Tuple[int, int, int]:
    """
    Get grayscale color.

    Parameters
    ----------
    white_to_black : float

    Returns
    -------
    color : Tuple

    Examples
    --------
    >>> get_color(0)
    (255, 255, 255)
    >>> get_color(0.5)
    (128, 128, 128)
    >>> get_color(1)
    (0, 0, 0)
    """
    if not (0 <= white_to_black <= 1):
        raise ValueError(
            f"white_to_black={white_to_black} is not in the interval [0, 1]"
        )

    index = 255 - int(255 * white_to_black)
    r, g, b = index, index, index
    return int(r), int(g), int(b)


def get_color_code(val: float, max_val: float) -> str:
    """
    Get a HTML color code which is between 0 and max_val.

    Parameters
    ----------
    val : number
    max_val : number

    Returns
    -------
    color_code : str

    Examples
    --------
    >>> get_color_code(0, 100)
    '#ffffff'
    >>> get_color_code(100, 100)
    '#000000'
    >>> get_color_code(50, 100)
    '#808080'
    """
    value = min(1.0, float(val) / max_val)
    r, g, b = get_color(value)
    return f"#{r:02x}{g:02x}{b:02x}"



"""Calculate the confusion matrix (CSV inputs)."""

# Core Library
import csv
import logging
from typing import List, Tuple

# Third party
import numpy as np
import numpy.typing as npt

# First party
import clana.io

logger = logging.getLogger(__name__)


def main(predictions_filepath: str, gt_filepath: str, n: int) -> None:
    """
    Calculate a confusion matrix.

    Parameters
    ----------
    predictions_filepath : str
        CSV file with delimter ; and quoting char "
        The first field is an identifier, the second one is the index of the
        predicted label
    gt_filepath : str
        CSV file with delimter ; and quoting char "
        The first field is an identifier, the second one is the index of the
        ground truth
    n : int
        Number of classes
    """
    # Read CSV files
    with open(predictions_filepath) as fp:
        reader = csv.reader(fp, delimiter=";", quotechar='"')
        predictions = [tuple(row) for row in reader]

    with open(gt_filepath) as fp:
        reader = csv.reader(fp, delimiter=";", quotechar='"')
        truths = [tuple(row) for row in reader]

    cm = calculate_cm(predictions, truths, n)
    path = "cm.json"
    clana.io.write_cm(path, cm)
    logger.info(f"cm was written to '{path}'")


def calculate_cm(
    truths: List[Tuple[str, ...]], predictions: List[Tuple[str, ...]], n: int
) -> npt.NDArray:
    """
    Calculate a confusion matrix.

    Parameters
    ----------
    truths : List[Tuple[str, str]]
    predictions : List[Tuple[str, str]]
    n : int
        Number of classes

    Returns
    -------
    confusion_matrix : numpy array (n x n)
    """
    cm = np.zeros((n, n), dtype=int)

    ident2truth_index = {}
    for identifier, truth_index in truths:
        ident2truth_index[identifier] = int(truth_index)

    if len(predictions) != len(truths):
        msg = f'len(predictions) = {len(predictions)} != {len(truths)} = len(truths)"'
        raise ValueError(msg)

    for ident, pred_index in predictions:
        cm[ident2truth_index[ident]][int(pred_index)] += 1

    return cm



"""Metrics for confusion matrices."""

# Third party
import numpy.typing as npt


def get_accuracy(cm: npt.NDArray) -> float:
    """
    Get the accuaracy by the confusion matrix cm.

    Parameters
    ----------
    cm : ndarray

    Returns
    -------
    accuracy : float

    Examples
    --------
    >>> import numpy as np
    >>> cm = np.array([[10, 20], [30, 40]])
    >>> get_accuracy(cm)
    0.5
    >>> cm = np.array([[20, 10], [30, 40]])
    >>> get_accuracy(cm)
    0.6
    """
    return float(sum(cm[i][i] for i in range(len(cm)))) / float(cm.sum())



"""Calculate the confusion matrix (one label per line)."""

# Core Library
import csv
import json
import logging
import os
import sys
from typing import Dict, List, Tuple

# Third party
import numpy as np
import numpy.typing as npt
import sklearn.metrics

# First party
import clana.utils

logger = logging.getLogger(__name__)


def main(
    label_filepath: str, gt_filepath: str, predictions_filepath: str, clean: bool
) -> None:
    """
    Get a simple confunsion matrix.

    Parameters
    ----------
    label_filepath : str
        Path to a CSV file with delimiter ;
    gt_filepath : str
        Path to a CSV file with delimiter ;
    predictions : str
        Path to a CSV file with delimiter ;
    clean : bool, optional (default: False)
        Remove classes that the classifier doesn't know
    """
    label_filepath = os.path.abspath(label_filepath)
    labels = clana.utils.load_labels(label_filepath, 0)

    # Read CSV files
    with open(gt_filepath) as fp:
        reader = csv.reader(fp, delimiter=";", quotechar='"')
        truths = [row[0] for row in reader]

    with open(predictions_filepath) as fp:
        reader = csv.reader(fp, delimiter=";", quotechar='"')
        predictions = [row[0] for row in reader]

    cm = calculate_cm(labels, truths, predictions, clean=False)
    # Write JSON file
    cm_filepath = os.path.abspath("cm.json")
    logger.info(f"Write results to '{cm_filepath}'.")
    with open(cm_filepath, "w") as outfile:
        str_ = json.dumps(
            cm.tolist(), indent=2, separators=(",", ": "), ensure_ascii=False
        )
        outfile.write(str_)
    print(cm)


def calculate_cm(
    labels: List[str],
    truths: List[str],
    predictions: List[str],
    replace_unk_preds: bool = False,
    clean: bool = False,
) -> npt.NDArray:
    """
    Calculate a confusion matrix.

    Parameters
    ----------
    labels : List[int]
    truths : List[int]
    predictions : List[int]
    replace_unk_preds : bool, optional (default: True)
        If a prediction is not in the labels in label_filepath, replace it
        with UNK
    clean : bool, optional (default: False)
        Remove classes that the classifier doesn't know

    Returns
    -------
    confusion_matrix : numpy array (n x n)
    """
    # Check data
    if len(predictions) != len(truths):
        msg = f"len(predictions) = {len(predictions)} != {len(truths)} = len(truths)"
        raise ValueError(msg)

    label2i = {}  # map a label to 0, ..., n
    for i, label in enumerate(labels):
        label2i[label] = i

    if clean:
        truths, predictions = clean_truths(truths, predictions)

    if replace_unk_preds:
        predictions = clean_preds(predictions, label2i)

    n = _sanity_check(truths, labels, label2i, predictions)

    # TODO: do no always filter
    filter_data_unk = True
    if filter_data_unk:
        truths2, predictions2 = [], []
        for tru, pred in zip(truths, predictions):
            if pred != "unk":  # TODO: tru != 'UNK'!!!
                truths2.append(tru)
                predictions2.append(pred)
        truths = truths2
        predictions = predictions2

    report = sklearn.metrics.classification_report(truths, predictions, labels=labels)
    print(report)
    print(f"Accuracy: {sklearn.metrics.accuracy_score(truths, predictions) * 100:.2f}%")

    cm = np.zeros((n, n), dtype=int)

    for truth_label, pred_label in zip(truths, predictions):
        cm[label2i[truth_label]][label2i[pred_label]] += 1

    return cm


def clean_truths(
    truths: List[str], predictions: List[str]
) -> Tuple[List[str], List[str]]:
    """
    Remove classes that the classifier doesn't know.

    Parameters
    ----------
    truths : List[int]
    predictions : List[int]

    Returns
    -------
    truths, predictions : List[int], List[int]
    """
    preds = []
    truths_tmp = []
    for tru, pred in zip(truths, predictions):
        if tru in predictions:
            truths_tmp.append(tru)
            preds.append(pred)
    predictions = preds
    truths = truths_tmp
    return truths, predictions


def clean_preds(predictions: List[str], label2i: Dict[str, int]) -> List[str]:
    """
    If a prediction is not in the labels in label_filepath, replace it with UNK.

    Parameters
    ----------
    predictions : List[str]
    label2i : Dict[str, int]
        Maps a label to an index

    Returns
    -------
    predictions : List[str]
    """
    preds = []
    for pred in predictions:
        if pred in label2i:
            preds.append(pred)
        else:
            preds.append("UNK")
    predictions = preds
    return predictions


def _sanity_check(
    truths: List[str],
    labels: List[str],
    label2i: Dict[str, int],
    predictions: List[str],
) -> int:
    for label in truths:
        if label not in label2i:
            logger.error(f"Could not find label '{label}'")
            sys.exit(-1)

    n = len(labels)
    for label in predictions:
        if label not in label2i:
            label2i[label] = len(labels)
            n = len(labels) + 1
            logger.error(
                f"Could not find label '{label}' in labels file => Add class UNK"
            )
    return n



"""Everything about clustering classes of a confusion matrix."""

# Core Library
import logging
import random
from typing import List, TypeVar, Union, cast

# Third party
import numpy as np
import numpy.typing as npt

# First party
import clana.utils

cfg = clana.utils.load_cfg()
logger = logging.getLogger(__name__)


T = TypeVar("T")


def apply_grouping(labels: List[T], grouping: List[int]) -> List[List[T]]:
    """
    Return list of grouped labels.

    Parameters
    ----------
    labels : List[T]
    grouping : List[int]

    Returns
    -------
    grouped_labels : List[List[T]]

    Examples
    --------
    >>> labels = ['de', 'en', 'fr']
    >>> grouping = [False, True]
    >>> apply_grouping(labels, grouping)
    [['de', 'en'], ['fr']]
    """
    groups = []
    current_group = [labels[0]]
    for label, cut in zip(labels[1:], grouping):
        if cut:
            groups.append(current_group)
            current_group = [label]
        else:
            current_group.append(label)
    groups.append(current_group)
    return groups


def _remove_single_element_groups(hierarchy: List[List[T]]) -> List[Union[T, List[T]]]:
    """
    Flatten sub-lists of length 1.

    Parameters
    ----------
    hierarchy : List[List]

    Returns
    -------
    hierarchy : list of el / lists

    Examples
    --------
    >>> hierarchy = [[0], [1, 2]]
    >>> _remove_single_element_groups(hierarchy)
    [0, [1, 2]]
    """
    h_new: List[Union[T, List[T]]] = []
    for el in hierarchy:
        if len(el) > 1:
            h_new.append(el)
        else:
            h_new.append(el[0])
    return h_new


def extract_clusters(
    cm: npt.NDArray,
    labels: List[str],
    steps: int = 10**4,
    lambda_: float = 0.013,
    method: str = "local-connectivity",
    interactive: bool = False,
) -> List[int]:
    """
    Find clusters in cm.

    Idea:
        mininmize lambda (error between clusters) - (count of clusters)
        s.t.: Each inter-cluster accuracy has to be lower than the overall
              accuracy

    Parameters
    ----------
    cm : npt.NDArray
    labels : List[str]
    steps : int
    lambda_ : float
        The closer to 0, the more groups
        The bigger, the bigger groups
    method : {'local-connectivity', 'energy'}
    interactive : bool

    Returns
    -------
    clustes : List[int]
    """
    if method == "energy":
        n = len(cm)
        grouping = [0 for _ in range(n - 1)]
        minimal_score = get_score(cm, grouping, lambda_)
        best_grouping = grouping[:]
        for _ in range(steps):
            pos = random.randint(0, n - 2)
            grouping = best_grouping[:]
            grouping[pos] = bool((grouping[pos] + 1) % 2)
            current_score = get_score(cm, grouping, lambda_)
            if current_score < minimal_score:
                best_grouping = grouping
                minimal_score = current_score
                logger.info(f"Best grouping: {grouping} (score: {minimal_score})")
    elif method == "local-connectivity":
        if interactive:
            thres: Union[float, int] = find_thres_interactive(cm, labels)
        else:
            thres = find_thres(cm, cfg["visualize"]["threshold"])
        logger.info(f"Found threshold for local connection: {thres}")
        best_grouping = split_at_con_thres(cm, thres, labels, interactive=interactive)
    else:
        raise NotImplementedError(f"method='{method}'")
    logger.info(f"Found {sum(best_grouping) + 1} clusters")
    return best_grouping


def create_weight_matrix(grouping: List[int]) -> npt.NDArray:
    """
    Create a matrix which contains the distance to the diagonal.

    Parameters
    ----------
    grouping : List[int]

    Returns
    -------
    weight_matrix : npt.NDArray
        A symmetric matrix
    """
    n = len(grouping) + 1
    weight_matrix = np.zeros((n, n))
    for i in range(n):
        seen_1 = False
        for j in range(i + 1, n):
            if seen_1:
                weight_matrix[i][j] = 1
            elif grouping[j - 1] == 1:
                seen_1 = True
                weight_matrix[i][j] = 1
    return weight_matrix + weight_matrix.transpose()


def get_score(cm: npt.NDArray, grouping: List[int], lambda_: float) -> float:
    """
    Get the score of a confusion matrix.

    Parameters
    ----------
    cm : npt.NDArray
    grouping : List[int]
    lambda_ : float

    Returns
    -------
    score : float
    """
    # First party
    from clana.visualize_cm import calculate_score

    inter_cluster_err = 0.0
    weights = create_weight_matrix(grouping)
    inter_cluster_err = calculate_score(cm, weights)
    return lambda_ * inter_cluster_err - sum(grouping)


def find_thres(cm: npt.NDArray, percentage: float) -> float:
    """
    Find a threshold for grouping.

    Parameters
    ----------
    cm : npt.NDArray
    percentage : float
        Probability that two neighboring classes belong togehter

    Returns
    -------
    connectivity : float
    """
    n = int(len(cm) * (1.0 - percentage)) - 1
    con = sorted(get_neighboring_connectivity(cm))
    return con[n]


def find_thres_interactive(cm: npt.NDArray, labels: List[str]) -> float:
    """
    Find a threshold for grouping.

    The threshold is the minimum connection strength for two classes to be
    within the same cluster.

    Parameters
    ----------
    cm : npt.NDArray
    labels : List[str]

    Returns
    -------
    pos_str : float
    """
    n = len(cm)
    con = sorted(zip(get_neighboring_connectivity(cm), zip(range(n - 1), range(1, n))))
    pos_str = None

    # Lowest position from which we know that they are connected
    pos_up = n - 1

    # Highest position from which we know that they are not connected
    neg_low = 0
    while pos_up - 1 > neg_low:
        print(f"pos_up={pos_up}, neg_low={neg_low}, pos_str={pos_str}")
        pos = int((pos_up + neg_low) / 2)
        con_str, (i1, i2) = con[pos]
        should_be_conn = input(
            f"Should {labels[i1]} and {labels[i2]} be in one cluster? (y/n): "
        )
        if should_be_conn == "n":
            neg_low = pos
        elif should_be_conn == "y":
            pos_up = pos
            pos_str = con_str
        else:
            print(f"Please type only 'y' or 'n'. You typed {should_be_conn}.")
    pos_str = cast(float, pos_str)
    return pos_str


def get_neighboring_connectivity(cm: npt.NDArray) -> List[float]:
    """
    Get how strong neighboring classes are connected.

    Parameters
    ----------
    cm : npt.NDArray

    Returns
    -------
    con : List[float]
    """
    con = []
    n = len(cm)
    for i in range(n - 1):
        con.append(cm[i][i + 1] + cm[i + 1][i])
    return con


def split_at_con_thres(
    cm: npt.NDArray, thres: float, labels: List[str], interactive: bool
) -> List[int]:
    """
    Two classes are not in the same group if they are not connected strong.

    Minimum connection strength is thres. The bigger this value, the more
    clusters / the smaller clusters you will get.
    """
    con = get_neighboring_connectivity(cm)
    grouping = []
    for i, el in enumerate(con):
        if el == thres and interactive:
            should_conn = "-"
            while should_conn not in ["y", "n"]:
                should_conn = input(
                    f"Should {labels[i]} and {labels[i + 1]} be in one "
                    "cluster? (y/n): "
                )
                if should_conn == "y":
                    grouping.append(0)
                elif should_conn == "n":
                    grouping.append(1)
                else:
                    print("please type either 'y' or 'n'")
        else:
            grouping.append(el < thres)
    return grouping
